@@ -1,4 +1,4 @@
 from . import architectures  # noqa: F401
-from .models.factory import create_model  # noqa: F401
+from .models.factory import create_model, create_preprocessing  # noqa: F401
 from .models.registry import list_models  # noqa: F401
 from .version import __version__  # noqa: F401
